@@ -6,31 +6,47 @@ Created by kevincar <kevincarrolldavis@gmail.com>
 
 """
 
+from __future__ import annotations
+
+import sys
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    if sys.platform != "darwin":
+        assert False, "This backend is only available on macOS"
+
 import asyncio
 import itertools
 import logging
-from typing import Any, Dict, Iterable, NewType, Optional
+from collections.abc import Iterable
+from typing import Any, Optional
+
+if sys.version_info < (3, 11):
+    from async_timeout import timeout as async_timeout
+else:
+    from asyncio import timeout as async_timeout
 
-import async_timeout
 import objc
-from Foundation import NSNumber, NSObject, NSArray, NSData, NSError, NSUUID, NSString
 from CoreBluetooth import (
+    CBUUID,
+    CBCharacteristic,
+    CBCharacteristicWriteType,
+    CBCharacteristicWriteWithResponse,
+    CBDescriptor,
     CBPeripheral,
     CBService,
-    CBCharacteristic,
-    CBDescriptor,
-    CBCharacteristicWriteWithResponse,
 )
+from Foundation import NSUUID, NSArray, NSData, NSError, NSObject
 
-from ...exc import BleakError
-from ..client import NotifyCallback
+from bleak.args.corebluetooth import NotificationDiscriminator
+from bleak.backends.client import NotifyCallback
+from bleak.exc import BleakError
 
 # logging.basicConfig(level=logging.DEBUG)
 logger = logging.getLogger(__name__)
 
-CBPeripheralDelegate = objc.protocolNamed("CBPeripheralDelegate")
 
-CBCharacteristicWriteType = NewType("CBCharacteristicWriteType", int)
+CBPeripheralDelegate = objc.protocolNamed("CBPeripheralDelegate")
 
 
 class PeripheralDelegate(NSObject):
@@ -38,7 +54,9 @@ class PeripheralDelegate(NSObject):
 
     ___pyobjc_protocols__ = [CBPeripheralDelegate]
 
-    def initWithPeripheral_(self, peripheral: CBPeripheral):
+    def initWithPeripheral_(
+        self, peripheral: CBPeripheral
+    ) -> Optional[PeripheralDelegate]:
         """macOS init function for NSObject"""
         self = objc.super(PeripheralDelegate, self).init()
 
@@ -51,24 +69,31 @@ class PeripheralDelegate(NSObject):
         self._event_loop = asyncio.get_running_loop()
         self._services_discovered_future = self._event_loop.create_future()
 
-        self._service_characteristic_discovered_futures: Dict[int, asyncio.Future] = {}
-        self._characteristic_descriptor_discover_futures: Dict[int, asyncio.Future] = {}
+        self._service_characteristic_discovered_futures: dict[
+            int, asyncio.Future[NSArray[CBCharacteristic]]
+        ] = {}
+        self._characteristic_descriptor_discover_futures: dict[
+            int, asyncio.Future[None]
+        ] = {}
 
-        self._characteristic_read_futures: Dict[int, asyncio.Future] = {}
-        self._characteristic_write_futures: Dict[int, asyncio.Future] = {}
+        self._characteristic_read_futures: dict[int, asyncio.Future[NSData]] = {}
+        self._characteristic_write_futures: dict[int, asyncio.Future[None]] = {}
 
-        self._descriptor_read_futures: Dict[int, asyncio.Future] = {}
-        self._descriptor_write_futures: Dict[int, asyncio.Future] = {}
+        self._descriptor_read_futures: dict[int, asyncio.Future[NSObject]] = {}
+        self._descriptor_write_futures: dict[int, asyncio.Future[None]] = {}
 
-        self._characteristic_notify_change_futures: Dict[int, asyncio.Future] = {}
-        self._characteristic_notify_callbacks: Dict[int, NotifyCallback] = {}
+        self._characteristic_notify_change_futures: dict[int, asyncio.Future[None]] = {}
+        self._characteristic_notify_callbacks: dict[int, NotifyCallback] = {}
+        self._characteristic_notification_discriminators: dict[
+            int, Optional[NotificationDiscriminator]
+        ] = {}
 
-        self._read_rssi_futures: Dict[NSUUID, asyncio.Future] = {}
+        self._read_rssi_futures: dict[NSUUID, asyncio.Future[int]] = {}
 
         return self
 
     @objc.python_method
-    def futures(self) -> Iterable[asyncio.Future]:
+    def futures(self) -> Iterable[asyncio.Future[Any]]:
         """
         Gets all futures for this delegate.
 
@@ -93,18 +118,22 @@ class PeripheralDelegate(NSObject):
         )
 
     @objc.python_method
-    async def discover_services(self) -> NSArray:
+    async def discover_services(
+        self, services: Optional[NSArray[CBUUID]] = None
+    ) -> NSArray[CBService]:
         future = self._event_loop.create_future()
 
         self._services_discovered_future = future
         try:
-            self.peripheral.discoverServices_(None)
+            self.peripheral.discoverServices_(services)
             return await future
         finally:
             del self._services_discovered_future
 
     @objc.python_method
-    async def discover_characteristics(self, service: CBService) -> NSArray:
+    async def discover_characteristics(
+        self, service: CBService
+    ) -> NSArray[CBCharacteristic]:
         future = self._event_loop.create_future()
 
         self._service_characteristic_discovered_futures[service.startHandle()] = future
@@ -115,12 +144,14 @@ class PeripheralDelegate(NSObject):
             del self._service_characteristic_discovered_futures[service.startHandle()]
 
     @objc.python_method
-    async def discover_descriptors(self, characteristic: CBCharacteristic) -> NSArray:
+    async def discover_descriptors(
+        self, characteristic: CBCharacteristic
+    ) -> NSArray[CBDescriptor]:
         future = self._event_loop.create_future()
 
-        self._characteristic_descriptor_discover_futures[
-            characteristic.handle()
-        ] = future
+        self._characteristic_descriptor_discover_futures[characteristic.handle()] = (
+            future
+        )
         try:
             self.peripheral.discoverDescriptorsForCharacteristic_(characteristic)
             await future
@@ -135,18 +166,20 @@ class PeripheralDelegate(NSObject):
     async def read_characteristic(
         self,
         characteristic: CBCharacteristic,
-        use_cached: bool = True,
+        use_cached: bool,
         timeout: int = 20,
     ) -> NSData:
-        if characteristic.value() is not None and use_cached:
-            return characteristic.value()
+        value = characteristic.value()
+        if value is not None and use_cached:
+            return value
 
         future = self._event_loop.create_future()
 
         self._characteristic_read_futures[characteristic.handle()] = future
+
         try:
             self.peripheral.readValueForCharacteristic_(characteristic)
-            async with async_timeout.timeout(timeout):
+            async with async_timeout(timeout):
                 return await future
         finally:
             del self._characteristic_read_futures[characteristic.handle()]
@@ -155,8 +188,9 @@ class PeripheralDelegate(NSObject):
     async def read_descriptor(
         self, descriptor: CBDescriptor, use_cached: bool = True
     ) -> Any:
-        if descriptor.value() is not None and use_cached:
-            return descriptor.value()
+        value = descriptor.value()
+        if value is not None and use_cached:
+            return value
 
         future = self._event_loop.create_future()
 
@@ -205,13 +239,19 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     async def start_notifications(
-        self, characteristic: CBCharacteristic, callback: NotifyCallback
+        self,
+        characteristic: CBCharacteristic,
+        callback: NotifyCallback,
+        notification_discriminator: Optional[NotificationDiscriminator] = None,
     ) -> None:
         c_handle = characteristic.handle()
         if c_handle in self._characteristic_notify_callbacks:
             raise ValueError("Characteristic notifications already started")
 
         self._characteristic_notify_callbacks[c_handle] = callback
+        self._characteristic_notification_discriminators[c_handle] = (
+            notification_discriminator
+        )
 
         future = self._event_loop.create_future()
 
@@ -238,9 +278,10 @@ class PeripheralDelegate(NSObject):
             del self._characteristic_notify_change_futures[c_handle]
 
         self._characteristic_notify_callbacks.pop(c_handle)
+        self._characteristic_notification_discriminators.pop(c_handle)
 
     @objc.python_method
-    async def read_rssi(self) -> NSNumber:
+    async def read_rssi(self) -> int:
         future = self._event_loop.create_future()
 
         self._read_rssi_futures[self.peripheral.identifier()] = future
@@ -254,7 +295,10 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     def did_discover_services(
-        self, peripheral: CBPeripheral, services: NSArray, error: Optional[NSError]
+        self,
+        peripheral: CBPeripheral,
+        services: NSArray[CBService],
+        error: Optional[NSError],
     ) -> None:
         future = self._services_discovered_future
         if error is not None:
@@ -280,9 +324,9 @@ class PeripheralDelegate(NSObject):
         self,
         peripheral: CBPeripheral,
         service: CBService,
-        characteristics: NSArray,
+        characteristics: NSArray[CBCharacteristic],
         error: Optional[NSError],
-    ):
+    ) -> None:
         future = self._service_characteristic_discovered_futures.get(
             service.startHandle()
         )
@@ -302,7 +346,7 @@ class PeripheralDelegate(NSObject):
 
     def peripheral_didDiscoverCharacteristicsForService_error_(
         self, peripheral: CBPeripheral, service: CBService, error: Optional[NSError]
-    ):
+    ) -> None:
         logger.debug("peripheral_didDiscoverCharacteristicsForService_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_discover_characteristics_for_service,
@@ -318,7 +362,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         future = self._characteristic_descriptor_discover_futures.get(
             characteristic.handle()
         )
@@ -341,7 +385,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didDiscoverDescriptorsForCharacteristic_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_discover_descriptors_for_characteristic,
@@ -355,21 +399,40 @@ class PeripheralDelegate(NSObject):
         self,
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
-        value: NSData,
+        value: Optional[NSData],
         error: Optional[NSError],
-    ):
+    ) -> None:
         c_handle = characteristic.handle()
 
         future = self._characteristic_read_futures.get(c_handle)
 
-        # If there is no pending read request, then this must be a notification
-        # (the same delagate callback is used by both).
-        if not future:
-            if error is None:
+        # If error is set, then we know this was a read response.
+        # Otherwise, if there is a pending read request, we can't tell if this is a read response or notification.
+        # If the user provided a notification discriminator, we can use that to
+        # identify if this callback is due to a notification by analyzing the value.
+        # If not, and there is a future (pending read request), we assume it is a read response but can't know for sure.
+        if not error:
+            assert value is not None
+
+            notification_discriminator = (
+                self._characteristic_notification_discriminators.get(c_handle)
+            )
+            if not future or (
+                notification_discriminator and notification_discriminator(bytes(value))
+            ):
                 notify_callback = self._characteristic_notify_callbacks.get(c_handle)
 
                 if notify_callback:
                     notify_callback(bytearray(value))
+                    return
+
+        if not future:
+            logger.warning(
+                "Unexpected event didUpdateValueForCharacteristic for 0x%04x with value: %r and error: %r",
+                c_handle,
+                value,
+                error,
+            )
             return
 
         if error is not None:
@@ -377,6 +440,7 @@ class PeripheralDelegate(NSObject):
             future.set_exception(exception)
         else:
             logger.debug("Read characteristic value")
+            assert value is not None
             future.set_result(value)
 
     def peripheral_didUpdateValueForCharacteristic_error_(
@@ -384,7 +448,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didUpdateValueForCharacteristic_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_update_value_for_characteristic,
@@ -399,9 +463,9 @@ class PeripheralDelegate(NSObject):
         self,
         peripheral: CBPeripheral,
         descriptor: CBDescriptor,
-        value: NSObject,
+        value: Optional[Any],
         error: Optional[NSError],
-    ):
+    ) -> None:
         future = self._descriptor_read_futures.get(descriptor.handle())
         if not future:
             logger.warning("Unexpected event didUpdateValueForDescriptor")
@@ -413,6 +477,7 @@ class PeripheralDelegate(NSObject):
             future.set_exception(exception)
         else:
             logger.debug("Read descriptor value")
+            assert value is not None
             future.set_result(value)
 
     def peripheral_didUpdateValueForDescriptor_error_(
@@ -420,7 +485,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         descriptor: CBDescriptor,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didUpdateValueForDescriptor_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_update_value_for_descriptor,
@@ -436,7 +501,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         future = self._characteristic_write_futures.get(characteristic.handle(), None)
         if not future:
             return  # event only expected on write with response
@@ -454,7 +519,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didWriteValueForCharacteristic_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_write_value_for_characteristic,
@@ -469,7 +534,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         descriptor: CBDescriptor,
         error: Optional[NSError],
-    ):
+    ) -> None:
         future = self._descriptor_write_futures.get(descriptor.handle())
         if not future:
             logger.warning("Unexpected event didWriteValueForDescriptor")
@@ -488,7 +553,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         descriptor: CBDescriptor,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didWriteValueForDescriptor_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_write_value_for_descriptor,
@@ -503,7 +568,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         c_handle = characteristic.handle()
         future = self._characteristic_notify_change_futures.get(c_handle)
         if not future:
@@ -525,7 +590,7 @@ class PeripheralDelegate(NSObject):
         peripheral: CBPeripheral,
         characteristic: CBCharacteristic,
         error: Optional[NSError],
-    ):
+    ) -> None:
         logger.debug("peripheral_didUpdateNotificationStateForCharacteristic_error_")
         self._event_loop.call_soon_threadsafe(
             self.did_update_notification_for_characteristic,
@@ -536,7 +601,7 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     def did_read_rssi(
-        self, peripheral: CBPeripheral, rssi: NSNumber, error: Optional[NSError]
+        self, peripheral: CBPeripheral, rssi: int, error: Optional[NSError]
     ) -> None:
         future = self._read_rssi_futures.get(peripheral.identifier(), None)
 
@@ -550,12 +615,21 @@ class PeripheralDelegate(NSObject):
         else:
             future.set_result(rssi)
 
-    # peripheral_didReadRSSI_error_ method is added dynamically later
+    def peripheral_didReadRSSI_error_(
+        self: PeripheralDelegate,
+        peripheral: CBPeripheral,
+        rssi: int,
+        error: Optional[NSError],
+    ) -> None:
+        logger.debug("peripheral_didReadRSSI_error_")
+        self._event_loop.call_soon_threadsafe(
+            self.did_read_rssi, peripheral, rssi, error
+        )
 
     # Bleak currently doesn't use the callbacks below other than for debug logging
 
     @objc.python_method
-    def did_update_name(self, peripheral: CBPeripheral, name: NSString) -> None:
+    def did_update_name(self, peripheral: CBPeripheral, name: str) -> None:
         logger.debug(f"name of {peripheral.identifier()} changed to {name}")
 
     def peripheralDidUpdateName_(self, peripheral: CBPeripheral) -> None:
@@ -566,55 +640,16 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     def did_modify_services(
-        self, peripheral: CBPeripheral, invalidated_services: NSArray
+        self, peripheral: CBPeripheral, invalidated_services: NSArray[CBService]
     ) -> None:
         logger.debug(
             f"{peripheral.identifier()} invalidated services: {invalidated_services}"
         )
 
     def peripheral_didModifyServices_(
-        self, peripheral: CBPeripheral, invalidatedServices: NSArray
+        self, peripheral: CBPeripheral, invalidatedServices: NSArray[CBService]
     ) -> None:
         logger.debug("peripheral_didModifyServices_")
         self._event_loop.call_soon_threadsafe(
             self.did_modify_services, peripheral, invalidatedServices
         )
-
-
-# peripheralDidUpdateRSSI:error: was deprecated and replaced with
-# peripheral:didReadRSSI:error: in macOS 10.13
-if objc.macos_available(10, 13):
-
-    def peripheral_didReadRSSI_error_(
-        self: PeripheralDelegate,
-        peripheral: CBPeripheral,
-        rssi: NSNumber,
-        error: Optional[NSError],
-    ) -> None:
-        logger.debug("peripheral_didReadRSSI_error_")
-        self._event_loop.call_soon_threadsafe(
-            self.did_read_rssi, peripheral, rssi, error
-        )
-
-    objc.classAddMethod(
-        PeripheralDelegate,
-        b"peripheral:didReadRSSI:error:",
-        peripheral_didReadRSSI_error_,
-    )
-
-
-else:
-
-    def peripheralDidUpdateRSSI_error_(
-        self: PeripheralDelegate, peripheral: CBPeripheral, error: Optional[NSError]
-    ) -> None:
-        logger.debug("peripheralDidUpdateRSSI_error_")
-        self._event_loop.call_soon_threadsafe(
-            self.did_read_rssi, peripheral, peripheral.RSSI(), error
-        )
-
-    objc.classAddMethod(
-        PeripheralDelegate,
-        b"peripheralDidUpdateRSSI:error:",
-        peripheralDidUpdateRSSI_error_,
-    )
